MU = 398600936839470.0  # Standard gravitational parameter for Earth (m^3/s^2)
RE = 6378136.3  # Earth's radius (m)
J2 = 1.08262668355315130e-3  # Earth's J2 coefficient
EARTH_ROTATION_RATE = 7.2921150e-5  # rad/s
#d